@@ -24,8 +24,10 @@ foreach(array('text', 'submit', 'hidden', 'password', 'reset', 'file',
     'color', 'range', 'button') as $t) 
   SDV($InputTags[$t][':html'], "<input type='$t' \$InputFormArgs />");
 
-foreach(array('text', 'password', 'email', 'url', 'tel', 'number', 'search', 'date', 'month') as $t) 
+foreach(array('text', 'textarea', 'password', 'email', 'url', 'tel', 'number', 'search', 'date', 'month') as $t) {
   SDV($InputTags[$t]['class'], "inputbox");
+  SDV($InputTags[$t][':args'], array('name', 'value', 'placeholder'));
+}
 
 foreach(array('submit', 'button', 'reset', 'color') as $t) 
   SDV($InputTags[$t]['class'], "inputbutton");
@@ -195,6 +197,28 @@ function InputToHTML($pagename, $type, $args, &$opt) {
     $dataattr = preg_grep('/^data-[-a-z]+$/', array_keys($opt));
     $attrlist = array_merge($attrlist, $dataattr);
   }
+  $datalist = '';
+  if (isset($opt['list'])) {
+    static $listcount = 0;
+    $list = MarkupRestore($opt['list']);
+    $alist = [];
+    if (strpos($list, "\n")!==false) {
+      $alist = preg_split("/\n+/", $list, -1, PREG_SPLIT_NO_EMPTY);
+    }
+    elseif (strpos($list, ";")!==false) {
+      $alist = preg_split("/; +/", $list, -1, PREG_SPLIT_NO_EMPTY);
+    }
+    if ($alist) {
+      $alist = array_map('trim', $alist);
+      $opt['list'] = 'pmdatalist'. (++$listcount);
+      $datalist = '';
+      $datalist = "<datalist id='{$opt['list']}'>";
+      foreach ($alist as $o) if($o>'') {
+        $datalist .= '<option value="'.PHSCQF($o).'"/>';
+      }
+      $datalist .= '</datalist>';
+    }
+  }
   $attr = array();
   foreach ($attrlist as $a) {
     if (!isset($opt[$a]) || $opt[$a]===false) continue;
@@ -204,7 +228,7 @@ function InputToHTML($pagename, $type, $args, &$opt) {
     $attr[] = "$a='".str_replace("'", '&#39;', $opt[$a])."'";
   }
   $FmtV['$InputFormArgs'] = implode(' ', $attr);
-  return FmtPageName($opt[':html'], $pagename);
+  return FmtPageName($opt[':html'], $pagename).$datalist;
 }
 
 
@@ -320,7 +344,7 @@ function RequestArgs($req = NULL) {
 ## Form-based authorization prompts (for use with PmWikiAuth)
 SDVA($InputTags['auth_form'], array(
   ':html' => "<form \$InputFormArgs>\$PostVars",
-  'action' => str_replace("'", '%37', stripmagic($_SERVER['REQUEST_URI'])),
+  'action' => str_replace("'", '%37', stripmagic($_SERVER['REQUEST_URI']??'')),
   'method' => 'post',
   'name' => 'authform'));
 SDV($AuthPromptFmt, array(&$PageStartFmt, 'page:$SiteGroup.AuthForm',
@@ -355,7 +379,7 @@ if (@$_REQUEST['editform']) {
   $PageEditForm=$_REQUEST['editform'];
   $PageEditFmt='$EditForm';
 }
-$Conditions['e_preview'] = '(boolean)$_REQUEST["preview"]';
+$Conditions['e_preview'] = '(bool)$_REQUEST["preview"]';
 
 # (:e_preview:) displays the preview of formatted text.
 Markup('e_preview', 'directives',
